from __future__ import annotations
import asyncio
import logging
import os
import threading
import websockets
import queue
from relationalai import Resources, debugging
from relationalai.clients.config import PUBLIC_CONFIG_KEYS, Config # to get the users root file path
import __main__

from relationalai.clients.profile_polling import ProfilePollerHandler

class DummyRecord(logging.LogRecord):
    def __init__(self, msg):
        self.msg = msg

#------------------------------------------------------------------------------
# Logging Handler
#------------------------------------------------------------------------------

class WebSocketLoggingHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from relationalai.analysis.mechanistic import Mechanism
        self.Mechanism = Mechanism
        self.queue: queue.Queue[str] = queue.Queue()

    def start(self, url: str, config: Config):
        main_path = getattr(__main__, "__file__", None)

        # Catch any spans that are the programs parent that were added before the debug client was started.
        open_spans = []
        existing_parent = debugging.TIP_SPAN
        while existing_parent:
            open_spans.append(existing_parent)
            existing_parent = existing_parent.parent
        open_spans.reverse()

        program_span = None

        for span in open_spans:
            if span.type == "program":
                program_span = span
            self.queue.put_nowait(self.format_raw({"event": "span_start", "span": span}))

        if not program_span:
            program_span = debugging.span_start(
                "program",
                main=os.path.relpath(main_path) if main_path else "notebook",
                **{k: config.get(k, None) for k in PUBLIC_CONFIG_KEYS if config.get(k, None) is not None}
            )

        self.thread = threading.Thread(target=self.connect, args=(url, program_span))
        self.thread.start()

    def emit(self, record):
        d = record.msg
        if (isinstance(d, dict) and
            d["event"] == "span_start" and
            "task" in d["span"].attrs and
            "mech" not in d["span"].attrs):
            d["span"].attrs["mech"] = self.Mechanism(d["span"].attrs["task"])
        log_entry = self.format(record)
        self.queue.put_nowait(log_entry)

    def format_raw(self, msg):
        return self.format(DummyRecord(msg))

    def connect(self, url: str, span: debugging.Span):
        try:
            try:
                loop = asyncio.get_running_loop()
                loop.run_until_complete(self.connect_async(url, span))
            except RuntimeError:
                asyncio.run(self.connect_async(url, span))
        except (ConnectionRefusedError, websockets.WebSocketException, OSError, asyncio.TimeoutError):
            pass

    async def connect_async(self, url: str, span: debugging.Span):
        async with websockets.connect(url) as ws:
            print(f"Connected to debugger at {url}...")
            alive = True
            while True:
                if not threading.main_thread().is_alive():
                    alive = False
                    debugging.span_flush()

                try:
                    log_entry = self.queue.get(timeout=1)
                    await ws.send(log_entry)
                except queue.Empty:
                    if not alive:
                        break
                except websockets.ConnectionClosedError:
                    break

#------------------------------------------------------------------------------
# Enable debugging
#------------------------------------------------------------------------------

already_debugging = False
def start_debugger_session(config: Config, host: str|None = None, port: int|None = None):
    global already_debugging
    if already_debugging:
        return

    if host is None:
        host = "0.0.0.0"

    if port is None:
        port = 8080

    ws_handler = WebSocketLoggingHandler()
    ws_handler.setFormatter(debugging.JsonFormatter())
    debugging.logger.addHandler(ws_handler)
    ws_handler.start(f"ws://{host}:{port}/ws/program", config)
    
    resources = Resources(config=config)
    debugging.logger.addHandler(ProfilePollerHandler(resources))

    already_debugging = True



from __future__ import annotations
import asyncio
import json
from pathlib import Path
import time
from typing import Any, Callable, Coroutine, Protocol, Union, cast
import webbrowser

import aiohttp
from websockets.exceptions import ConnectionClosed
from aiohttp import web

import relationalai
from relationalai.clients.client import ResourcesBase
from relationalai.debugging import encode_log_message

MAX_MSG_SIZE = 4*4194304 # 16MB
PROJECT_ROOT = Path(relationalai.__file__).resolve()

# This is weird and dynamic because the frontend directory is in different
# places depending on whether you're running with `-e`, or install a package.
max_ancestor_steps = 3
FRONTEND_PATH = PROJECT_ROOT
while not (FRONTEND_PATH / "frontend").exists() and max_ancestor_steps > 0:
    FRONTEND_PATH = FRONTEND_PATH.parent
    max_ancestor_steps -= 1

HTTP_DIST_DIR = FRONTEND_PATH / "frontend" / "debugger" / "dist"

# The maximum number of messages to be sent at once when catching a new client up
MAX_BATCH_SIZE = 500

_resources: ResourcesBase|None = None

def get_resources() -> ResourcesBase:
    if not _resources:
        raise Exception("Unable to acquire resources")

    return _resources

#------------------------------------------------------------------------------
# Program Connections
#------------------------------------------------------------------------------

connected_program = None

async def handle_program(req: web.Request):
    ws = web.WebSocketResponse(max_msg_size=MAX_MSG_SIZE)
    await ws.prepare(req)

    global connected_program
    if connected_program:
        print("Refusing to connect to new program until previous one finishes.")
        await ws.close()

    connected_program = ws
    print("Program connected")
    try:
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                await broadcast(msg.data)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                err = ws.exception()
                if err:
                    if "Decompressed message size" in str(err):
                        await broadcast_error("Failed to send large message from program", err)
                    else:
                        raise err
            else:
                raise Exception("Unknown message type", msg.type)
    finally:
        print("Program disconnected")
        connected_program = None

    return ws

#------------------------------------------------------------------------------
# Browser Client Connections
#------------------------------------------------------------------------------

connected_clients: set[web.WebSocketResponse] = set()

async def handle_client(req: web.Request):
    ws = web.WebSocketResponse(max_msg_size=MAX_MSG_SIZE)
    await ws.prepare(req)
    connected_clients.add(ws)
    try:
        for i in range(0, len(buffered_broadcasts), MAX_BATCH_SIZE):
            chunk = ",".join(buffered_broadcasts[i:i + MAX_BATCH_SIZE])
            await ws.send_str(f"[{chunk}]")

        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                await handle_client_action(msg.data, ws)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                err = ws.exception()
                if err:
                    raise err
            else:
                raise Exception("Unknown message type", msg.type)

    except ConnectionClosed as e:
        print(f"Client disconnected with reason: {e}")
    finally:
        connected_clients.remove(ws)

    return ws

class NoSuchActionError(KeyError):
    def __init__(self, action: str|None):
        super().__init__(f"Got action request from client for non-existent action of type '{action}'")

async def handle_client_action(msg: Any, ws: web.WebSocketResponse):
    parsed = None
    try:
        parsed = json.loads(msg)
        action_type = parsed.get("type")
        action = client_actions.get(action_type)
        if not action:
            raise NoSuchActionError(action_type)
        kwargs = {k: v for k, v in parsed.items() if k != "type" and k != "rpc_id"}
        res = action(ws, **kwargs) or {}

    except Exception as e:
        print("Error:", str(e))
        res = {"error": str(e)}

    if parsed:
        if asyncio.iscoroutine(res):
            res = await res

        res = cast(dict, res)
        res["event"] = "rpc_response"
        res["rpc_id"] = parsed.get("rpc_id")

        def dump_json(v):
            return json.dumps(v, default=encode_log_message)

        await ws.send_json([res], None, dumps=dump_json)

#------------------------------------------------------------------------------
# Browser Client Actions
#------------------------------------------------------------------------------

MaybeDict = Union[dict, None]
ClientActionReturn = Union[Coroutine[Any, Any, MaybeDict], MaybeDict]

class ClientAction(Protocol):
    def __call__(self, ws: web.WebSocketResponse, **kwargs: Any) -> ClientActionReturn:
        ...

client_actions: dict[str, ClientAction] = {}

def client_action(name: str|None = None):
    def decorator(fn: Callable[..., ClientActionReturn]):
        client_actions[name or fn.__name__] = fn

    return decorator

@client_action()
def clear(_):
    buffered_broadcasts.clear()

@client_action()
def list_transactions(_, **kwargs):
    return {"transactions": get_resources().list_transactions(**kwargs)}

@client_action()
def list_imports(_, **kwargs):
    if get_resources().platform == "snowflake":
        imports = get_resources().list_imports(**kwargs)
    else:
        imports = []
    return {"imports": imports}

@client_action()
def get_imports_status(_, **kwargs):
    return {"imports_status": get_resources().get_imports_status(**kwargs)}

#------------------------------------------------------------------------------
# Event Log
#------------------------------------------------------------------------------

buffered_broadcasts = []

async def broadcast(message):
    buffered_broadcasts.append(message)
    if connected_clients:
        tasks = [asyncio.create_task(client.send_str(f"[{message}]" )) for client in connected_clients]
        await asyncio.wait(tasks)

def broadcast_error(msg: str, err: BaseException):
    print(f"[WARN] {msg}")
    return broadcast(json.dumps({"event": "error", "origin": "debug-server", "message": msg, "cause": str(err), "stack": str(err.__traceback__)}))

#------------------------------------------------------------------------------
# Server
#------------------------------------------------------------------------------

async def serve_index(_):
    return web.FileResponse(HTTP_DIST_DIR / "index.html")

runner: web.AppRunner|None = None
async def run_server(host: str, port: int):
    global runner
    if runner:
        await runner.shutdown()

    app = web.Application()
    app.router.add_get("/ws/program", handle_program)
    app.router.add_get("/ws/client", handle_client)
    app.router.add_get("/", serve_index)
    app.router.add_static("/", HTTP_DIST_DIR, show_index=False)  # Serving static files from DIR
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, host, port)
    await site.start()

    url = f"http://{host}:{port}"
    time.sleep(0.25)
    print(f"Server started at {url}")
    webbrowser.open(url)

    return runner

def start_server(host: str, port: int|None = None, profile:str|None = None):
    global _resources
    _resources = relationalai.Resources(profile)

    if port is None:
        try:
            config_port = _resources.config.get("debug.port", 8080)
            if isinstance(config_port, int):
                port = config_port
            else:
                raise Exception("Invalid value specified for `debug.port`, expected `int`.")
        except AttributeError:
            port = 8080

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_server(host, port))
    loop.run_forever()


def stop_server():
    global runner
    if not runner:
        return

    loop = asyncio.get_running_loop()
    loop.run_until_complete(broadcast(json.dumps({"event": "debug_server_closed"})))
    loop.run_until_complete(runner.shutdown())
    runner = None



#--------------------------------------------------
# FLAGS
#--------------------------------------------------
USE_GRAPH_INDEX = True

#--------------------------------------------------
# Constants
#--------------------------------------------------

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

class GlobalProfileSetting:
    def __init__(self):
        self.profile = None

    def set(self, profile):
        self.profile = profile

    def get(self):
        return self.profile

GlobalProfile = GlobalProfileSetting()

SNOWFLAKE = "Snowflake"
AZURE = "Azure (Beta)"

SNOWFLAKE_AUTHENTICATOR = {
    "USERNAME & PASSWORD": "snowflake",
    "USERNAME & PASSWORD (MFA ENABLED)": "username_password_mfa",
    "SINGLE SIGN-ON (SSO)": "externalbrowser",
}

FIELD_PLACEHOLDER = ""

SNOWFLAKE_PROFILE_DEFAULTS = {
    "platform": "snowflake",
    "user": FIELD_PLACEHOLDER,
    "account": FIELD_PLACEHOLDER,
    "role": "PUBLIC",
    "warehouse": FIELD_PLACEHOLDER,
    "rai_app_name": FIELD_PLACEHOLDER,
    "engine": FIELD_PLACEHOLDER,
    "engine_size": "HIGHMEM_X64_S"
}

SNOWFLAKE_AUTHS = {
    "snowflake": {
        "authenticator": "snowflake",
        "password": FIELD_PLACEHOLDER,
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
    "username_password_mfa": {
        "authenticator": "username_password_mfa",
        "password": FIELD_PLACEHOLDER,
        "passcode": FIELD_PLACEHOLDER,
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
    "externalbrowser": {
        "authenticator": "externalbrowser",
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
}

AZURE_ENVS = {
    "Production": {
        "host": "azure.relationalai.com",
        "client_credentials_url": "https://login.relationalai.com/oauth/token"
    },
    "Early Access": {
        "host": "azure-ea.relationalai.com",
        "client_credentials_url": "https://login-ea.relationalai.com/oauth/token"
    },
    "Staging": {
        "host": "azure-staging.relationalai.com",
        "client_credentials_url": "https://login-staging.relationalai.com/oauth/token"
    },
    "Latest": {
        "host": "azure-latest.relationalai.com",
        "client_credentials_url": "https://login-latest.relationalai.com/oauth/token"
    },
}



from __future__ import annotations
import sys
from types import TracebackType
import ast
import io
import re
import textwrap
from typing import Any, List, Literal, NamedTuple, Tuple

from rich.console import Console
from rich.text import Text
from rich.table import Table
from rich import box
from enum import Enum
import contextlib
import contextvars

from .metamodel import Action, Task
from .environments import runtime_env, IPythonEnvironment, SnowbookEnvironment
from . import debugging


#--------------------------------------------------
# Print helpers
#--------------------------------------------------


_current_console = contextvars.ContextVar('current_console', default=None)

def get_console(*args, **kwargs):
   current = _current_console.get()
   if current is not None:
       return current
   return Console(*args, **kwargs)

@contextlib.contextmanager
def using_console(console):
   token = _current_console.set(console)
   try:
       yield
   finally:
       _current_console.reset(token)

@contextlib.contextmanager
def record(retain_buffer = False):
    buffer = io.StringIO()
    console = Console(record=True, file=buffer)
    with using_console(console):
        try:
            yield (console, buffer)
        finally:
            if not retain_buffer:
                buffer.close()
       
def rich_str(item:Any, style:str|None = None) -> str:
    output = io.StringIO()
    console = Console(file=output, force_terminal=True)
    console.print(item, style=style)
    return output.getvalue()

def body_text(console, body:str):
    body = textwrap.dedent(body)
    for line in body.splitlines():
        if not line.startswith("  "):
            console.print(line)
        else:
            console.print(line, soft_wrap=True)

def mark_source(source: debugging.SourceInfo|None, start_line:int|None=None, end_line:int|None=None, indent=8, highlight="yellow", highlight_lines = []):
    if source is None:
        return ""

    final_lines = []
    all_lines = source.source.splitlines()
    source_start = source.source_start_line or source.line or start_line or 0
    start_line = start_line if start_line is not None else source.line
    max_line = source.line + len(all_lines)
    end_line = end_line if end_line is not None else max_line
    line_number_len = len(str(max_line))
    for ix, line in enumerate(source.source.splitlines()):
        cur = line
        cur_indent = indent if ix > 0 else 0
        line_number = source_start + ix
        color = "dim white"
        if (line_number >= start_line and line_number <= end_line) or line_number in highlight_lines:
            color = highlight + " bold"
        cur = f"{' '*cur_indent}[{color}]  {line_number :>{line_number_len}} |  {cur}[/{color}]"
        final_lines.append(cur)
    return "\n".join(final_lines)

def print_source_error(source, name:str, content:str, color="red"):
    fixed_content_length = len(name) + len(source.file) + len(str(source.line)) + 2  # 2 for the spaces around the dash
    num_dashes = 74 - fixed_content_length
    dashes = '-' * num_dashes
    console = get_console(width=80, force_jupyter=False, stderr=True)
    console.print("\n")
    console.print(f"[{color}]--- {name} {dashes} {source.file}: {source.line}")
    console.print()
    body_text(console, content)
    console.print()
    console.print(f'[{color}]{"-" * 80}')
    console.print()

def print_error(name:str, content:str, color="red"):
    fixed_content_length = len(name) + 2  # 2 for the spaces around the dash
    num_dashes = 76 - fixed_content_length
    dashes = '-' * num_dashes
    console = get_console(width=80, force_jupyter=False, stderr=True)
    console.print("\n")
    console.print(f"[{color}]--- {name} {dashes}")
    console.print()
    body_text(console, content)
    console.print()
    console.print(f'[{color}]{"-" * 80}')
    console.print()

def print_error_name(name:str, color="red"):
    console = get_console(width=80, force_jupyter=False, stderr=True)
    console.print("\n")
    console.print(f'[{color}]{"-" * 80}')
    console.print()
    body_text(console, name)
    console.print()
    console.print(f'[{color}]{"-" * 80}')
    console.print()

SuggestionEnv = Literal["cli", "notebook", "python"]
class Suggestion(NamedTuple):
   env: SuggestionEnv
   cmd: str
   style: str|None = "green"

def suggest(suggestions: list[Suggestion], prefix = "with the following command", prefix_multi: str|None = None):
   """Format a list of suggestions, adapted to the current runtime environment."""
   if prefix_multi is None:
      prefix_multi = prefix + "s"

   relevant: list[Suggestion] = []
   for suggestion in suggestions:
      relevant += adapt_to_env(suggestion)

   if len(relevant) == 1:
      return "\n".join([
         prefix + ":",
         "",
         fmt_suggestion(relevant[0]),
      ])
   else:
      return "\n".join([
         prefix_multi + ":",
         "",
         "\n\n".join(fmt_suggestion(suggestion) for suggestion in relevant)
      ])

def fmt_suggestion(suggestion: Suggestion) -> str:
   """Format a single styled suggestion."""
   cmd = textwrap.dedent(suggestion.cmd)
   return "\n".join(f"[{suggestion.style}]{line}[/{suggestion.style}]" for line in cmd.splitlines()) if suggestion.style else cmd

def adapt_to_env(suggestion: Suggestion) -> list[Suggestion]:
   """Adapt a suggestion to work in the current runtime environment if possible, or exclude it if not."""
   if suggestion.env == "cli":
      if isinstance(runtime_env, SnowbookEnvironment) and runtime_env.runner == "warehouse":
         return []
      if isinstance(runtime_env, (IPythonEnvironment, SnowbookEnvironment)):
         return [Suggestion("notebook", "\n".join(f"! {subcmd}" for subcmd in partition_by_indent(textwrap.dedent(suggestion.cmd))), suggestion.style)]
      return [suggestion]
   else:
      return [suggestion]

def partition_by_indent(text):
    # Split on newlines followed by non-whitespace
    groups = re.split(r'\n(?=\S)', text)
    # Strip any trailing whitespace from each group
    return [group.rstrip() for group in groups]

#--------------------------------------------------
# Transformers
#--------------------------------------------------

class IfToWithTransformer(ast.NodeTransformer):
    def visit_If(self, node):
        with_node = ast.With(
            items=[ast.withitem(context_expr=node.test, optional_vars=None)],
            body=node.body,
            lineno=node.lineno,
            type_comment=None)
        return with_node

class WithDynamic(ast.NodeTransformer):
    def visit_With(self, node):
        content = ast.unparse(node.items[0].context_expr).replace(")", "dynamic=True)")
        with_node = ast.With(
            items=[ast.withitem(context_expr=ast.Name(id=content), optional_vars=node.items[0].optional_vars)],
            body=[],
            lineno=node.lineno,
            type_comment=None)
        return with_node

class SetToMethod(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.targets[0], ast.Attribute):
            keyword = ast.keyword(arg=node.targets[0].attr, value=node.value)
            return ast.Expr(value=ast.Call(
                func=ast.Attribute(value=node.targets[0].value, attr="set", ctx=ast.Load()),
                args=[],
                keywords=[keyword],
                lineno=node.lineno
            ))
        return node

class AssignToCompare(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.targets[0], ast.Attribute) and len(node.targets) == 1:
            compare_node = ast.Compare(
                left=node.targets[0],
                ops=[ast.Eq()],
                comparators=[node.value]
            )
            expr_node = ast.Expr(value=compare_node)
            ast.copy_location(expr_node, node)

            return expr_node
        return node

class PropertyNameReplacer(ast.NodeTransformer):
    def __init__(self, old_name, new_name):
        self.old_name = old_name
        self.new_name = new_name

    def visit_Attribute(self, node):
        # Check if the attribute name matches the old name
        if isinstance(node.attr, str) and node.attr == self.old_name:
            # Replace the attribute name with the new name
            node.attr = self.new_name
        return node

    def visit_Name(self, node):
        # Check if the variable name matches the old name
        if isinstance(node.id, str) and node.id == self.old_name:
            # Replace the variable name with the new name
            node.id = self.new_name
        return node

#--------------------------------------------------
# Finders
#--------------------------------------------------

class PropertyFinder(ast.NodeVisitor):
    def __init__(self, start_line, properties):
        self.errors = []
        self.start_line = start_line
        self.properties = properties
        self.found_properties_lines = []  # To store lines where properties are found
        self.dynamic_properties = []  # To store dynamic properties

    def to_line_numbers(self, node):
        return (node.lineno, node.end_lineno)

    def visit_Attribute(self, node):
        if node.attr in self.properties:
            line_numbers = self.to_line_numbers(node)
            if line_numbers[0] >= self.start_line:
                self.found_properties_lines.append(node.lineno)
        self.generic_visit(node)

    def visit_Call(self, node):
        # Check if this is a call to 'getattr'
        if (isinstance(node.func, ast.Name) and node.func.id == 'getattr' and
                len(node.args) >= 2):
            if isinstance(node.args[1], ast.Str):
                property_name = node.args[1].s
                if property_name in self.properties:
                    line_numbers = self.to_line_numbers(node)
                    if line_numbers[0] >= self.start_line:
                        self.found_properties_lines.append(node.lineno)
            else:
                line_numbers = self.to_line_numbers(node)
                if line_numbers[0] >= self.start_line:
                    self.dynamic_properties.append(node.lineno)
        self.generic_visit(node)

#--------------------------------------------------
# Metaclass
#--------------------------------------------------

class RAIPostInitMeta(type):
    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        instance.__post_init__()
        return instance

# --------------------------------------------------
# RAIException
# --------------------------------------------------

class RAIException(Exception, metaclass=RAIPostInitMeta):
    def __init__(self, message, name=None, content=None, source: debugging.SourceInfo|None=None, **kwargs):
        super().__init__(message)
        self.name = name
        self.message = message
        self.content = content
        self.source = source

        # Store any additional keyword arguments as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return self.message

    def __post_init__(self):
        self.raw_content = self.strip_rich_tags()

    def pprint(self):
        if self.source and self.content:
            print_source_error(self.source, self.name or self.message, self.content.strip())
        elif self.content:
            print_error(self.name or self.message, self.content.strip())
        else:
            print_error_name(self.name or self.message)

    def __eq__(self, other):
        if isinstance(other, RAIException):
            return self.message == other.message and self.__dict__ == other.__dict__
        return False

    def strip_rich_tags(self):
        if self.content:
            # Use Text.from_markup to remove rich tags
            plain_text = Text.from_markup(self.content).plain
            # Split the content into lines
            lines = plain_text.split('\n')
            # Remove leading empty lines or lines with only whitespace
            while lines and not lines[0].strip():
                lines.pop(0)
            # Strip leading whitespace from the first non-empty line
            if lines:
                lines[0] = lines[0].lstrip()
            # Join the lines back together
            cleaned_content = '\n'.join(lines)
            return cleaned_content
        else:
            return self.message

    def clone(self):
        try:
            raise AssertionError
        except AssertionError:
            traceback = sys.exc_info()[2]
            back_frame = traceback.tb_frame.f_back if traceback else None

            if back_frame is None:
                return self.with_traceback(traceback)

        back_tb = TracebackType(tb_next=None,
                                    tb_frame=back_frame,
                                    tb_lasti=back_frame.f_lasti,
                                    tb_lineno=back_frame.f_lineno)
        return self.with_traceback(back_tb)

    def message_for_environment(self, message, cli_command, provider_call):
        if isinstance(runtime_env, SnowbookEnvironment):
            return textwrap.dedent(f"""
            {message} with the following call:

            [green]{provider_call}[/green]
            """)
        else:
            return textwrap.dedent(f"""
            {message} with either of the following commands:

            [green]{cli_command}[/green]

            [green]{provider_call}[/green]
            """)

class ModelError(RAIException):
    error_locations = {}

    def __init__(self, problems):
        super().__init__("Error object added")
        problem = problems[0]
        self.problems = problems
        self.source = ModelError.error_locations.get(problem["props"]["pyrel_id"], debugging.SourceInfo())
        self.message = problem["message"]
        self.name = "Model error"
        self.content = self.format_message()

    def get_formatted_props_string(self):
        # Create a table
        table = Table(show_header=True, box=box.ROUNDED, padding=(0, 1))
        ks = list([k for k in self.problems[0]["props"].keys() if not k.startswith("pyrel_")])
        for k in ks:
            table.add_column(k)

        for problem in self.problems:
            row = []
            for k in ks:
                row.append(str(problem["props"].get(k, "")))
            table.add_row(*row)

        with io.StringIO() as string_io:
            console = Console(file=string_io)
            console.print(table)
            table_string = string_io.getvalue()

        return table_string

    def format_message(self):
        row_str = self.get_formatted_props_string()

        marked = mark_source(self.source, self.source.line, self.source.line, indent=2) if self.source else ""
        return "\n\n".join([self.message, "  " + marked, row_str])


class RelQueryError(RAIException):
    def __init__(self, problem, source: debugging.SourceInfo|None = None):
        super().__init__("Rel query error")
        self.problem = problem
        self.source = source
        self.message = "Query error"
        self.name = "Query error"
        self.content = self.format_message()

    def format_message(self):
        problem = self.problem
        marked = mark_source(self.source, -1, -1) if self.source else ""
        return textwrap.dedent(f"""
        {problem["report"]}
        {problem["message"]}

        {marked}
        """)

class RAIExceptionSet(RAIException):
    def __init__(self, exceptions: List[RAIException]):
        super().__init__("Multiple Errors, see above")
        self.exceptions = exceptions

    def pprint(self):
        for exception in self.exceptions:
            exception.pprint()

    def clone(self):
        return RAIExceptionSet([exception.clone() for exception in self.exceptions])

# --------------------------------------------------
# Warning
# --------------------------------------------------

class RAIWarning(Warning, metaclass=RAIPostInitMeta):
    name = ""
    message: str|None = None
    content = None
    source = None

    def __init__(self, message="", name=None, content=None, source=None):
        self.name = name
        self.message = message
        self.content = content
        self.source = source

    def pprint(self):
        msg = self.name or self.message or ""
        if self.source and self.content:
            print_source_error(self.source, msg, self.content.strip(), color="yellow")
        elif self.content:
            print_error(msg, self.content.strip(), color="yellow")
        else:
            print_error_name(msg, color="yellow")

    def __post_init__(self):
        self.raw_content = self.strip_rich_tags()
        debugging.warn(self)

    def strip_rich_tags(self):
        if self.content:
            # Use Text.from_markup to remove rich tags
            plain_text = Text.from_markup(self.content).plain
            # Split the content into lines
            lines = plain_text.split('\n')
            # Remove leading empty lines or lines with only whitespace
            while lines and not lines[0].strip():
                lines.pop(0)
            # Strip leading whitespace from the first non-empty line
            if lines:
                lines[0] = lines[0].lstrip()
            # Join the lines back together
            cleaned_content = '\n'.join(lines)

            return cleaned_content
        else:
            return self.name

    def __str__(self):
        return f"{self.name}: {self.raw_content}"

class RelQueryWarning(RAIWarning):
    def __init__(self, problem, source):
        super().__init__("Query warning")
        self.name = "Query warning"
        self.problem = problem
        self.source = source
        self.message = "Query warning"
        self.content = self.format_message()

    def format_message(self):
        problem = self.problem
        marked = mark_source(self.source, -1, -1)
        return textwrap.dedent(f"""
        {problem["report"]}
        {problem["message"]}

        {marked}
        """)

class UnknownSourceWarning(RAIWarning):
    def __init__(self, source_fqns: list[str]):
        name = "Unknown Source"
        if len(source_fqns) == 1:
            msg = textwrap.fill(f"Unable to access source '{source_fqns[0]}'. Ensure the source's fully qualified name is spelled correctly and that the role your program is executing with has permission to access it.".strip(), 78)
        else:
            prefix = "The following sources could not be accessed:"
            suffix = textwrap.fill("Ensure the sources' fully qualified names are spelled correctly and that the role your program is executing with has permission to access them.", 78)

            msg = "\n".join([
                prefix,
                "",
                *[f"- {fqn}" for fqn in source_fqns],
                "",
                suffix
            ])

        super().__init__(name, name, msg)
        self.source_fqns = source_fqns

class InvalidSourceTypeWarning(RAIWarning):
    def __init__(self, source_types: dict[str, str]):
        if len(source_types) == 1:
            [(fqn, type)] = source_types.items()
            msg = f"RAI currently only supports tables and views.\n The source '{fqn}' has type '{type}'"
        else:
            prefix = "RAI currently only supports tables and views. These have other types:"
            msg = "\n".join([
                prefix,
                "",
                *[f"- {fqn} ({type})" for fqn, type in source_types.items()]
            ])

        name = "Invalid source type"
        super().__init__(name, name, msg)
        self.source_types = source_types

class IntegrityConstraintViolation(RAIException):
    def __init__(self, violation, source=None):
        self.violation = violation
        self.source = source

        body = violation["message"]
        self.name = "Integrity constraint violation"
        self.message = "Integrity constraint violation"
        self.content = self.format_message(body)
        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self, body):
        marked = mark_source(self.source, -1, -1)

        f"Integrity constraint violation\n{body}"

        return textwrap.dedent(f"""
        {body}

        {marked}
        """)

    def clone(self):
        return IntegrityConstraintViolation(self.violation, self.source)

class ModelWarning(RAIWarning):
    error_locations = {}

    def __init__(self, problems):
        super().__init__("Error object added")
        problem = problems[0]
        self.problems = problems
        self.source = ModelError.error_locations.get(problem["props"]["pyrel_id"], debugging.SourceInfo())
        self.message = problem["message"]
        self.name = "Model warning"
        self.content = self.format_message()

    def get_formatted_props_string(self):
        # Create a table
        table = Table(show_header=True, box=box.ROUNDED, padding=(0, 1))
        ks = list([k for k in self.problems[0]["props"].keys() if not k.startswith("pyrel_")])
        for k in ks:
            table.add_column(k)

        for problem in self.problems:
            row = []
            for k in ks:
                row.append(problem["props"].get(k, ""))
            table.add_row(*row)

        with io.StringIO() as string_io:
            console = Console(file=string_io)
            console.print(table)
            table_string = string_io.getvalue()

        return table_string

    def format_message(self):
        row_str = self.get_formatted_props_string()

        marked = mark_source(self.source, self.source.line, self.source.line, indent=2) if self.source else ""
        return "\n\n".join([self.message or "", "  " + marked, row_str])


# --------------------------------------------------
# Rel Exceptions/warnings
# --------------------------------------------------

class NumericOverflow(RAIException):
    def __init__(self, problem, source):
        super().__init__("Numeric Overflow")
        self.problem = problem
        self.source = source
        self.message = "Numeric Overflow"
        self.name = "Numeric Overflow"
        self.content = self.format_message()

    def format_message(self):
        problem = self.problem
        marked = mark_source(self.source, -1, -1)
        info = problem["report"].split("\n--------------")[0].strip()
        return textwrap.dedent(f"""
        {info}

        {marked}
        """)

class ArityMismatch(RAIWarning):
    def __init__(self, problem, source):
        self.name = "Arity mismatch"
        self.message = problem["message"]
        self.source = source
        self.content = self.format_message()

    def format_message(self):
        source = self.source
        match = re.search(r"`(.+?)` expects (.*)", self.message or "")
        if match is None:
            return self.message
        fq_name = match.group(1)
        rel_name = fq_name.split("::")[-1]
        arity_sentence = match.group(2)
        found = PropertyFinder(source.line, [rel_name])
        if source.block:
            found.visit(source.block)
        found_lines = found.found_properties_lines or found.dynamic_properties
        marked = mark_source(self.source, -1, -1, highlight_lines=found_lines)
        return textwrap.dedent(f"""
        The relation [yellow]{rel_name}[/yellow] expects {arity_sentence}

        {marked}
        """)

class UninitializedPropertyException(RAIException):
    def __init__(self, undefineds: List[Tuple[str, Any]]):
        self.content = ""
        self.raw_content = ""
        self.source = None
        self.name = "Uninitialized property"
        self.message = "Uninitialized property"

        # group by source
        source_map: dict[str, tuple[debugging.SourceInfo, list[str]]] = {}
        for name, source in undefineds:
            if source.source not in source_map:
                source_map[source.source] = (source, [name])
            else:
                source_map[source.source][1].append(name)

        chunks: list[str] = []
        for (source, names) in source_map.values():
            self.source = source

            unique_names = list(set(names))
            props = ", ".join([f"[yellow]{name}[/yellow]" for name in unique_names])
            prop_line = (
                f"property {props} has"
                if len(unique_names) == 1
                else f"properties {props} have"
            )
            found = PropertyFinder(source.line, unique_names)
            if source.block:
                found.visit(source.block)
            found_lines = found.found_properties_lines or found.dynamic_properties
            marked = mark_source(source, -1, -1, indent=12, highlight_lines=found_lines)
            chunks.append(textwrap.dedent(f"""
            The {prop_line} never been set or added to and so will always cause the rule or query to fail.

            {marked}
            """))

            self.content = "\n\n".join(chunks)

        super().__init__(self.name, self.message, self.content, self.source)


class RAITypeError(RAIException, TypeError):
    def __init__(self, param: str, message: str):
        self.param = param
        self.message = message
        self.name = "Invalid type error"
        self.source = Errors.call_source(4)

        # Mark the source if available
        if self.source:
            self.marked = mark_source(self.source, self.source.line, self.source.line)
        else:
            self.marked = ""

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        if self.source:
            return textwrap.dedent(f"""
            The parameter [yellow]{self.param}[/yellow] is of the wrong type.

            {self.marked}
            """)
        else:
            return textwrap.dedent(f"""
            The parameter [yellow]{self.param}[/yellow] is of the wrong type.
            """)

class RAIValueError(RAIException, ValueError):
    def __init__(self, message: str):
        self.message = message
        self.name = "Invalid value error"
        self.source = Errors.call_source(4)

        # Mark the source if available
        if self.source:
            self.marked = mark_source(self.source, self.source.line, self.source.line)
        else:
            self.marked = ""

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        if self.source:
            return textwrap.dedent(f"""
            {self.message}

            {self.marked}
            """)
        else:
            return textwrap.dedent(f"""
            {self.message}
            """)


# --------------------------------------------------
# DSL scope errors
# --------------------------------------------------
class OutOfContextException(RAIException):
    def __init__(self, source=None):
        self.name = "Outside of context"
        self.message = (
            "Looks like this object is being used outside of a rule or query."
        )
        self.source = source or Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Looks like this [yellow]object[/yellow] is being used outside of a rule or query.

        {marked}
        """)


class VariableOutOfContextException(RAIException):
    def __init__(self, source, name: str|None, is_property=False):
        self.name = "Variable out of context"
        if name is None:
            self.message = "Looks like a variable is being used outside of the rule or query it was defined in."
        else:
            self.message = f"Looks like a variable representing '{name}' is being used outside of the rule or query it was defined in."
        self.source = source
        self.is_property = is_property
        self.content = self.format_message(name)

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self, name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Looks like a variable representing [yellow bold]{name}[/yellow bold] is being used outside of the rule or query it was defined in.

        {marked}
        """)

class SelectOutOfContext(RAIException):
    def __init__(self):
        self.name = "Outside of context"
        self.message = (
            "Looks like this select is being used outside of the query it was originally from."
        )
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Looks like this [yellow]select[/yellow] is being used outside of the query it is from.

        {marked}
        """)

#--------------------------------------------------
# DSL Type errors
#--------------------------------------------------

class FilterAsValue(RAIException):
    def __init__(self, source=None):
        self.name = "Filter used as a value"
        self.message = "Boolean expressions are filters and can't be used as values directly, use std.as_bool to cast them into True/False."
        self.source = source or Errors.call_source()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line) if self.source else ""
        return textwrap.dedent(f"""
        Boolean expressions are filters and can't be used as values directly. You can use relationalai.std.as_bool(..) to cast them into True/False.

        {marked}
        """)

class AsBoolForNonFilter(RAIException):
    def __init__(self, source=None):
        self.name = "as_bool used on non-filter"
        self.message = "as_bool can only be used on boolean expressions."
        self.source = source or Errors.call_source()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line) if self.source else ""
        return textwrap.dedent(f"""
        as_bool can only be used on boolean expressions.

        {marked}
        """)

class PropertyCaseMismatch(RAIWarning):
    def __init__(self, name, found_name):
        self.name = "Similar property name"
        self.message = ".{name} doesn't exist, but .{found_name} does"
        self.source = Errors.call_source()
        self.content = self.format_message(name, found_name)

    def format_message(self, name, found_name):
        source = self.source or debugging.SourceInfo()
        found = PropertyFinder(source.line, [name])
        if source.block:
            found.visit(source.block)
        found_lines = found.found_properties_lines or found.dynamic_properties
        marked = mark_source(source, -1, -1, indent=8, highlight_lines=found_lines)

        updated = mark_source(
            source.modify(PropertyNameReplacer(name, found_name)),
            -1, -1, indent=8, highlight_lines=found_lines, highlight="green",
        )

        return textwrap.dedent(f"""
        The property [yellow]{name}[/yellow] doesn't exist, but the very similar [green]{found_name}[/green] does.

        {marked}

        Did you mean [green]{found_name}[/green] instead?

        {updated}
        """)


class NonVarObject(RAIException, TypeError):
    def __init__(self, obj: Any, message: str):
        self.obj = obj
        self.message = message
        self.name = "non-var object used as variable"
        self.source = Errors.call_source(4)

        # Mark the source if available
        if self.source:
            self.marked = mark_source(self.source, self.source.line, self.source.line)
        else:
            self.marked = ""

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        if self.source:
            return textwrap.dedent(f"""
            [yellow]{self.obj.__class__.__name__}[/yellow] objects can't be used as variables, such as in a .add(), .set(), or select().

            {self.marked}
            """)
        else:
            return textwrap.dedent(f"""
            [yellow]{self.obj.__class__.__name__}[/yellow] objects can't be used as variables, such as in a .add(), .set(), or select().
            """)

class KeyedCantBeExtended(RAIException):
    def __init__(self):
        self.name = "Keyed types can't be extended"
        self.message = "Keyed types can't be extended."
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Keyed types can't be extended, use .add() in a rule instead.

        {marked}
        """)

class KeyedWrongArity(RAIException):
    def __init__(self, name:str, keys:List[str], given:int):
        self.name = "Incorrect number of keys"
        verb = "was" if given == 1 else "were"
        self.message = f"{name} expects {len(keys)} keys ({', '.join(keys)}), but only {given} {verb} given."
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        {self.message}

        {marked}
        """)

class RowLiteralTooLargeWarning(RAIWarning):
    def __init__(self, size, max_size=1000):
        self.name = "Large row literal"
        self.message = f"Rows with more than {max_size} items can be slow to compile."
        self.size = size
        self.max_size = max_size
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Rows with more than {self.max_size} items can be slow to compile. This one has {self.size} items:

        {marked}
        """)

class RowLiteralTooLarge(RAIException):
    def __init__(self, size, max_size=10000):
        self.name = "Row literal too large"
        self.message = f"Rows can't have more than {max_size} items."
        self.size = size
        self.max_size = max_size
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Rows can't have more than {self.max_size} items. This one has {self.size} items:

        {marked}
        """)

class RowLiteralMismatch(RAIException):
    def __init__(self, mismatched_thing:str):
        self.name = "Row literal mismatch"
        self.message = f"All rows in a row literal must have the same {mismatched_thing}"
        self.mismatched_thing = mismatched_thing
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        All rows in a row literal must have the same {self.mismatched_thing}

        {marked}
        """)


# --------------------------------------------------
# DSL property exceptions
# --------------------------------------------------


class ReservedPropertyException(RAIException):
    def __init__(self, source, property_name: str):
        self.name = "Reserved property name"
        self.message = f"The property '{property_name}' is a reserved property name on RelationalAI types."
        self.source = source
        self.content = self.format_message(property_name)

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self, property_name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The property '{property_name}' is a reserved property name on RelationalAI types.

        {marked}
        """)


class NonCallablePropertyException(RAIException):
    def __init__(self, source, property_name: str):
        self.name = "Non-callable property"
        self.message = f"The property '{property_name}' is not callable."
        self.source = source
        self.content = self.format_message(property_name)

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self, property_name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The property '{property_name}' is not callable on RelationalAI types.

        {marked}
        """)

class InvalidPropertySetException(RAIException):
    def __init__(self, source):
        self.name = "Invalid property set"
        self.message = "You can't set properties directly on a RAI object."
        self.source = source
        self.start_line = source.line
        self.end_line = source.line
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()
        super().__init__(
            message=self.message, name=self.name, source=source, content=self.content
        )

    def format_message(self):
        marked = self.mark_source()
        dynamic = mark_source(
            self.source.modify(SetToMethod()),
            self.start_line,
            self.end_line,
            highlight="green",
        )
        compare = mark_source(
            self.source.modify(AssignToCompare()),
            self.start_line,
            self.end_line,
            highlight="green",
        )

        return textwrap.dedent(f"""
        You can't set properties directly on a RAI object.

        {marked}

        If you are trying to set the value of a property use [green]set()[/green]:

        {dynamic}

        Or maybe you meant [green]==[/green] instead?

        {compare}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)

class MultipleIdentities(RAIException):
    def __init__(self):
        source = Errors.call_source()
        self.name = "Multiple identities"
        self.message = "You can't pass multiple identity variables for a single object."
        self.source = source
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line) if self.source else ""
        return textwrap.dedent(f"""
        You can't pass multiple identity variables for a single object.

        {marked}
        """)

# --------------------------------------------------
# Graph library exceptions
# --------------------------------------------------

class DirectedGraphNotApplicable(RAIValueError):
    def __init__(self, name: str):
        super().__init__(name)
        self.name = f"algorithm `{name}` is not applicable to directed graphs"

class DirectedGraphNotSupported(RAIValueError):
    def __init__(self, name: str):
        super().__init__(name)
        self.name = f"algorithm `{name}` does not currently support directed graphs"


class ParameterTypeMismatch(RAIValueError):
    def __init__(self, name: str, type_, value):
        super().__init__(name)
        self.name = (
            f"parameter `{name}` must be of type {type_.__name__.lower()}, "
            f"but its value {value!r} is of type {type(value)}"
        )

class ParameterBoundBelowInclusive(RAIValueError):
    def __init__(self, name: str, value, minimum):
        super().__init__(name)
        self.name = f"parameter `{name}` must be greater than or equal to {minimum}, but is {value!r}"

class ParameterBoundAboveInclusive(RAIValueError):
    def __init__(self, name: str, value, maximum):
        super().__init__(name)
        self.name = f"parameter `{name}` must be less than or equal to {maximum}, but is {value!r}"

class ParameterBoundBelowExclusive(RAIValueError):
    def __init__(self, name: str, value, minimum):
        super().__init__(name)
        self.name = f"parameter `{name}` must be strictly greater than {minimum}, but is {value!r}"

class ParameterBoundAboveExclusive(RAIValueError):
    def __init__(self, name: str, value, maximum):
        super().__init__(name)
        self.name = f"parameter `{name}` must be strictly less than {maximum}, but is {value!r}"


# --------------------------------------------------
# Engine exceptions
# --------------------------------------------------


class EngineNotFoundException(RAIException):
    def __init__(self, engine_name: str, message: str):
        self.engine_name = engine_name
        self.message = message
        self.name = "Engine unavailable"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        message = "We were unable to detect an existing engine or provision a new engine for you. You can manually create an engine"
        return self.message_for_environment(
            message, 
            f"rai engines:create --name {self.engine_name}",
            f"rai.Provider().create_engine('{self.engine_name}')"
        )

class EngineAutoCreateFailed(RAIException):
    def __init__(self, engine_name: str):
        self.engine_name = engine_name
        self.message = "Engine auto-creation failed"
        self.name = "Engine auto-creation failed"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        message = f"We tried to auto-create the engine [yellow]{self.engine_name}[/yellow]but it failed. You can manually create the engine"
        return self.message_for_environment(
            message,
            f"rai engines:create --name {self.engine_name}",
            f"rai.Provider().create_engine('{self.engine_name}')"
        )

class EnginePending(RAIException):
    def __init__(self, engine_name: str):
        self.engine_name = engine_name
        self.message = "Engine is in a pending state"
        self.name = "Engine not ready"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        message = f"The engine [yellow]{self.engine_name}[/yellow] is in a pending state. You can see the status of engines"
        return self.message_for_environment(
            message,
            "rai engines:list",
            "rai.Provider().list_engines()"
        )


class EngineResumeFailed(RAIException):
    def __init__(self, engine_name: str):
        self.engine_name = engine_name
        self.message = "Unable to resume engine"
        self.name = "Unable to resume engine"
        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()
        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        suggestions = [
           Suggestion("cli", f"""
           rai engines:delete --name {self.engine_name}
           rai engines:create --name {self.engine_name}
           """[1:-1]),
           Suggestion("python", f"""
           rai.Provider().delete_engine('{self.engine_name}')
           rai.Provider().create_engine('{self.engine_name}')
           """[1:-1])
        ]
        return "\n".join([
           f"The engine [yellow]{self.engine_name}[/yellow] could not be automatically resumed. Please ensure both your native app install and the `relationalai` package are up to date.",
           "",
           f"You can also manually recreate the engine {suggest(suggestions)}"
        ])

# --------------------------------------------------
# Hex exceptions
# --------------------------------------------------

class HexSessionException(RAIException):
    def __init__(self):
        self.name = "Hex session error"
        self.message = "Missing Snowflake session object in Hex"
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent("""
        To use RelationalAI in Hex, get your Snowflake session and supply it as a connection parameter to `rai.Model` or `rai.Provider`:

        [green]
        import hextoolkit
        hex_snowflake_conn = hextoolkit.get_data_connection('<Your Connection Name>')
        hex_snowpark_session = hex_snowflake_conn.get_snowpark_session()
        [/green]
        """)


# --------------------------------------------------
# Snowflake errors
# --------------------------------------------------

access_integration_warning = textwrap.dedent("""
Make sure that S3_RAI_INTERNAL_BUCKET_EGRESS_INTEGRATION is enabled. You can access the toggle from the vertical ellipsis button in the top-right corner of the Snowflake notebook window: ⋮ > Notebook settings > External access.

If you don't see S3_RAI_INTERNAL_BUCKET_EGRESS_INTEGRATION listed, get an account administrator to run Step 4 in the installation notebook here:

https://relational.ai/notebooks/relationalai-installation.ipynb
""")

class SnowflakeIntegrationMissingException(RAIException):
    def __init__(self):
        self.message = access_integration_warning
        self.name = "Integration not enabled"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return access_integration_warning

class SnowflakeAppMissingException(RAIException):
    def __init__(self, app_name: str):
        self.app_name = app_name
        self.message = f"Either the app '{app_name}' isn't installed in this Snowflake account or you don't have permission to access it."
        self.name = "Couldn't find RelationalAI Snowflake application"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        Either the app '{self.app_name}' isn't installed in this Snowflake account or you don't have permission to access it.

        If it's installed under a different name, run 'rai init' on the command line and you can set the app name.

        If the app is installed but your current role doesn't have permission to access the app, switch to a role that does.
        """)


class SnowflakeImportMissingException(RAIException):
    def __init__(self, source: debugging.SourceInfo|None, import_name: str, model_name: str):
        self.source = source
        self.import_name = import_name
        self.model_name = model_name
        self.message = (
            f"The Snowflake object '{import_name}' hasn't been imported into RAI."
        )
        self.name = "Couldn't find import"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        message = f"The Snowflake object '{self.import_name}' hasn't been imported into RAI. You can create an import for it using"
        return self.message_for_environment(
            message,
            f"rai imports:create --source {self.import_name} --model {self.model_name}",
            f"rai.Provider().create_streams(['{self.import_name}'], '{self.model_name}')"
        )


class SnowflakeChangeTrackingNotEnabledException(RAIException):
    def __init__(self, obj: tuple[str, str] | list[tuple[str, str]]):
        self.obj = obj
        if isinstance(obj, tuple):
            fqn, type = obj
            self.sql = f"[green]ALTER {type.upper()} {fqn} SET CHANGE_TRACKING = TRUE;[/green]"
            self.message = f"Change tracking isn't enabled for '{obj}'."
        elif isinstance(obj, list):
            self.sql = "\n".join([f"[green]ALTER {type.upper()} {fqn} SET CHANGE_TRACKING = TRUE;[/green]" for fqn, type in obj])
            self.message = f"Change tracking isn't enabled for the following sources: {', '.join(fqn for fqn, _ in obj)}"

        self.name = "Change tracking not enabled"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        if isinstance(self.obj, tuple):
            fqn_str, _ = self.obj
        elif isinstance(self.obj, list):
            fqn_str = ', '.join(fqn for fqn, _ in self.obj)
        else:
            fqn_str = self.obj
        return (
            f"Change tracking isn't enabled for [yellow]{fqn_str}[/yellow].\n\n"
            f"To enable change tracking, you'll need to run the following SQL:\n\n"
            f"{self.sql}"
        ).strip()

class SnowflakeTableObject:
    def __init__(self, message: str, source: str):
        self.message = message
        self.source = source

class SnowflakeTableObjectsException(RAIException):
    def __init__(self, table_objects: List[SnowflakeTableObject]):
        self.table_objects = table_objects
        table_count = len(table_objects)
        table_word = "table" if table_count == 1 else "tables"

        self.message = f"Getting the following {table_word} failed with the error in Snowflake."
        self.name = f"Snowflake {table_word.capitalize()} Streaming Error"

        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        formatted_objects = [
            f"Source: [yellow bold]{obj.source}[/yellow bold]\n\n"
            f"Message: {obj.message.strip()}"
            for obj in self.table_objects
        ]
        return "\n\n".join(formatted_objects)

class SnowflakeInvalidSource(RAIException):
    def __init__(self, source, type_source: str):
        self.source = source
        self.type_source = type_source
        self.message = "Invalid source"
        self.name = "Invalid source"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The source provided isn't a fully qualified Snowflake table or view name:

        {marked}
        """)


class SnowflakeRaiAppNotStarted(RAIException):
    def __init__(self, app_name: str):
        self.app_name = app_name
        self.message = f"The RelationalAI app '{app_name}' isn't started."
        self.name = "The RelationalAI application not started"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        Your app '{self.app_name}' has not been activated.
        You can activate it by running the following SQL:

        [green]call {self.app_name}.app.activate();[/green]
        """)


# --------------------------------------------------
# RAI Warnings
# --------------------------------------------------

class InvalidIfWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid if"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()

    def format_message(self):
        marked = self.mark_source()
        updated = mark_source(
            self.source.modify(IfToWithTransformer()),
            self.start_line,
            self.end_line,
            highlight="green",
        ) if self.source else ""
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        ) if self.source else ""

        return textwrap.dedent(f"""
        In a RelationalAI query, using an if statement dynamically modifies the structure of the query itself, rather than adding a conditional.

        {marked}

        If you're trying to do an action based on a condition, use a [green]with[/green] statement instead.

        {updated}

        If you are trying to create a dynamic query where parts are conditional, add the [green]dynamic=True[/green] flag to the query like so:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidLoopWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid loop"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()

    def format_message(self):
        marked = self.mark_source() if self.source else ""
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        ) if self.source else ""

        return textwrap.dedent(f"""
        In a RelationalAI query, using a loop statement would dynamically modify the query itself, like a macro.

        {marked}

        If that's the goal, you can add the [green]dynamic=True[/green] flag to the query:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidTryWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid try"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()

    def format_message(self):
        marked = self.mark_source()
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        ) if self.source else ""

        return textwrap.dedent(f"""
        In a RelationalAI query, using a try statement will have no effect unless a macro-like function is being called and can fail.

        {marked}

        If macro-like behavior is the goal, you can add the [green]dynamic=True[/green] flag to the query:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidBoolWarning(RAIWarning):
    def __init__(self, source):
        self.name = "Invalid boolean expression with Producer"
        self.source = source
        self.start_line = source.line
        self.end_line = source.line
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

    def format_message(self):
        marked = self.mark_source()
        return textwrap.dedent(f"""
        In a RelationalAI query, the truth values of Producer objects are unknown
        until the query has been evaluated. You may not use Producers in boolean
        expressions, such as those involving [green]if[/green], [green]while[/green], [green]and[/green], [green]or[/green], and [green]not[/green]:

        {marked}

        Producer objects include:

        - [green]Instance[/green] objects, such as [green]person[/green].
        - [green]InstanceProperty[/green] objects, such as [green]person.age[/green].
        - [green]Expression[/green] objects, such as [green]person.age >= 18[/green].
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)

# whether the python library or the Rel schema is out of date (or both!)
class InvalidVersionKind(Enum):
    LibraryOutOfDate = 'LibraryOutOfDate'
    SchemaOutOfDate = 'SchemaOutOfDate'
    Incompatible = 'Incompatible'

class RAIInvalidVersionWarning(RAIWarning):
    def __init__(
        self,
        kind: InvalidVersionKind,
        expected: List[Tuple[str, str, str]],
        lock: dict,
        platform: str,
        model_name: str,
        app_name: str,
        engine_name: str,
        errors: List[Tuple[InvalidVersionKind, str, str]] = []
    ):
        super().__init__()
        self.libraries = self.get_current_libs(errors)
        self.expected_lib_version = self.get_expected_lib_version(expected)
        self.current_lib_version = self.get_current_lib_version(lock)
        self.model_name = model_name
        self.app_name = app_name
        self.engine_name = engine_name

        self.name = "Invalid version"

        if kind == InvalidVersionKind.LibraryOutOfDate:
            msg = textwrap.dedent("""
                The RelationalAI Python library is out of date with respect to the schema. You can update the library using the following command:

                [green]pip install relationalai --upgrade[/green]
            """).strip()
        elif kind == InvalidVersionKind.SchemaOutOfDate:
            schema_fix = textwrap.dedent(f"""
                You can update the schema using the following stored procedure:

                {self.generate_command_line()}
                """).strip()

            if platform == "azure":
                schema_fix = "Unfortunately, the schema must be recreated in Azure."
            schema_msg = textwrap.dedent("""
                The schema is out of date with respect to the RelationalAI library.
                """).strip()
            msg = f"{schema_msg} {schema_fix}"
        else:
            command = "pip install relationalai --upgrade"
            warehouse_notebook = (
                isinstance(runtime_env, SnowbookEnvironment)
                and runtime_env.runner == "warehouse"
            )
            if isinstance(runtime_env, IPythonEnvironment) and not warehouse_notebook:
                command = "!" + command
            if warehouse_notebook:
                lib_msg = textwrap.dedent("""
                    The RelationalAI Python library is incompatible with the current schema.

                    You can update the library by downloading the latest ZIP file from https://relational.ai/relationalai.zip and uploading it here.
                """).strip()
            else:
                lib_msg = textwrap.dedent(f"""
                    The RelationalAI Python library is incompatible with the current schema.
                    You can update the library using the following command:

                    [green]{command}[/green]
                """).strip()
            if platform == "azure":
                schema_fix = "In addition, the schema must be recreated in Azure."
            else:
                schema_fix = textwrap.dedent(f"""
                In addition, you can update the schema using the following stored procedure:

                {self.generate_command_line()}
                """).strip()
            msg = f"{lib_msg}\n\n{schema_fix}"

        if len(self.libraries) > 0:
            library_info = self.generate_library_info()
            full_msg = f"{msg}\n\n{library_info}"
        else:
            full_msg = msg

        self.content = full_msg

    def extract_version(self, version_obj):
        match = re.search(r"\('([^']+)'\)", repr(version_obj))
        if match:
            return match.group(1)
        else:
            return ""

    def get_expected_lib_version(self, expected):
        libraries = {}
        for name, _, version_range in expected:
            from_version = self.extract_version(version_range[0])
            to_version = self.extract_version(version_range[1])
            libraries[name] = {"from": from_version, "to": to_version}
        return libraries

    def get_current_lib_version(self, lock):
        libraries = {}
        for k, v in lock.items():
            libraries[k[0]] = v
        return libraries

    def get_current_libs(self, errors=[]):
        if len(errors) == 0:
            return []

        libraries = []
        for _, library, _ in errors:
            libraries.append(library)
        return libraries

    def generate_library_info(self):
        if len(self.libraries) == 1:
            lib = self.libraries[0]
            current_version = self.current_lib_version.get(lib, "unknown")
            expected_version = self.expected_lib_version.get(lib, {})
            from_version = expected_version.get("from", "unknown")
            to_version = expected_version.get("to", "unknown")
            return (
                f"Your current library '{lib}' has version {current_version}. "
                f"However, it is expected to have a version range between {from_version} and {to_version}."
            )
        else:
            info = "Your current libraries have the following issues:\n"
            for lib in self.libraries:
                current_version = self.current_lib_version.get(lib, "unknown")
                expected_version = self.expected_lib_version.get(lib, {})
                from_version = expected_version.get("from", "unknown")
                to_version = expected_version.get("to", "unknown")
                info += (
                    f"- Library '{lib}' has version {current_version}. "
                    f"Expected version range: {from_version} to {to_version}.\n"
                )
            return info.strip()

    def generate_command_line(self):
        return f"[green]CALL {self.app_name or '<relationalai_app_name>'}.api.update_libraries('{self.model_name or '<model_name>'}', '{self.engine_name or '<engine_name>'}')[/green]"

class UnsupportedVisualizationError(RAIException):
    def __init__(self):
        name = "Unsupported Visualization Error"
        self.source = Errors.call_source() or debugging.SourceInfo()
        super().__init__(name, name, self.format_message(), self.source)

    def format_message(self):
        marked = mark_source(self.source)
        return textwrap.dedent(f"""
        This environment does not support embedding interactive visualizations via Graph.visualize().

        {marked}

        You can use graph.fetch() [1] to retrieve the necessary data to render your own visualization via whatever means the environment supports. Alternatively, consider running your program in another supported environment like a python script or jupyter.

        [1]: https://relational.ai/docs/reference/python/std/graphs/Graph/fetch
        """)

class EngineSizeMismatchWarning(RAIWarning):
    def __init__(self, engine_name: str, existing_size: str, requested_size: str):
        self.engine_name = engine_name
        self.existing_size = existing_size
        self.requested_size = requested_size
        self.message = f"An engine named '{engine_name}' already exists with a different size ('{existing_size}'). The existing engine will be used."
        self.name = "Engine size mismatch"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        An engine named '{self.engine_name}' already exists with a different size ('{self.existing_size}'). The existing engine will be used. You can change the size of '{self.engine_name}' by deleting the existing engine.

        You can delete the engine with either of the following commands:
        
        [green]rai engines:delete --name {self.engine_name}[/green]

        [green]rai.Provider().delete_engine('{self.engine_name}')[/green]
        """)


#--------------------------------------------------
# Deprecations + Migrations
#--------------------------------------------------

class DeprecationWarning(RAIWarning):
    def __init__(self, api: str, fix: str|None = None):
        self.api = api
        self.fix = fix

        self.name = "Deprecation Warning"
        self.prefix = f"'{api}' is deprecated and should no longer be used."
        self.message =  f"{self.prefix} {fix}" if fix else self.prefix
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        {self.prefix}

        {marked}

        {self.fix}
        """)

class SnowflakeProxyAPIDeprecationWarning(DeprecationWarning):
   def __init__(self):
      super().__init__("clients.snowflake.Snowflake", "Instead, use the `source` keyword argument when defining types. See https://relational.ai/docs/reference/python/Model/Type for details.")

class SnowflakeProxySourceError(RAIException):
    def __init__(self):
        self.name = "Snowflake Proxy Source Error"
        self.message =  "Data sources derived from the deprecated snowflake proxy 'clients.snowflake.Snowflake' aren't supported with your current configuration."
        self.source = Errors.call_source() or debugging.SourceInfo()
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        {self.message}

        {marked}

        Instead, use the `source` keyword argument when defining types. See https://relational.ai/docs/reference/python/Model/Type for details.
        """)


#TODO: Take out call_source from Errors class and put it in the main module
class Errors:
    @staticmethod
    def call_source(steps=None):
        if steps is None:
            return runtime_env.get_source()
        return runtime_env.get_source(steps + 1)




TOP_LEVEL_PROFILE_NAME = "__top_level__"
PARTIAL_PROFILE_NAME = "__partial__"
DEFAULT_PROFILE_NAME = "default"

#region Spans

# For a given span_type, the keys that are allowed to be inserted in the span's attrs
SPAN_ATTR_ALLOW_LIST = {
    'transaction': {'txn_id'},
    'wait': {'txn_id'},
}

# Don't insert these as attrs since they get their own columns
SPAN_FILTER_ATTRS = {
    "event",
    "span",
    "id",
    "parent_id",
    "start_time",
    "start_timestamp",
    "end_time",
    "end_timestamp",
    "elapsed",
}

# For a given span type, the attribute to use as its key.
SPAN_TYPES_KEYS = {
    'test': 'name',
    'benchmark': 'name',
    'run': 'idx',
    'query': 'tag',
    'test_session': 'platform', # 'platform' is how the cloud_provider is stored in the config
    'get_model': 'name',
    'rule': 'name',
}

#endregion



from __future__ import annotations
import io
import logging
import random
import os
from pathlib import Path
import textwrap
from typing import Callable, TypedDict

from pandas import DataFrame
import pytest

import rich
from rich.syntax import Syntax

import relationalai as rai
from relationalai.clients import config as cfg
from relationalai import debugging
from relationalai.debugging import logger
from relationalai.errors import RAIException
from relationalai.metamodel import Task

################################################################################
# Parameterize tests over all files in a directory
################################################################################

def path_to_slug(path: Path, base_path:str|Path):
    return str(path.relative_to(base_path)).replace("/", "__").replace(".py", "")

def all_files_in_dir(test_case_dir: Path, *, pattern = "*.py"):
    "Decorator to parameterize a pytest over files in the given directory."
    test_case_files = [path for path in test_case_dir.rglob(pattern)]
    return pytest.mark.parametrize(
        "file_path", test_case_files, ids=lambda path: path_to_slug(path, test_case_dir)
    )

################################################################################
# Tap block info out of the debug log
################################################################################

class ISourceInfo(TypedDict):
    file: str
    line: int
    block: str

class IPassInfo(TypedDict):
    name: str
    task: str
    elapsed: float

class ISampledResults(TypedDict):
    values: DataFrame
    count: int

class Block(TypedDict):
    task: Task
    source: ISourceInfo
    passes: list[IPassInfo]
    emitted: str
    emit_time: float
    results: ISampledResults|None
    elapsed: float|None

class QueryTextLogger(logging.Handler):
    def __init__(self):
        super().__init__()
        # compilation events that get 'results' added to them
        self.blocks: list[Block] = []

    def emit(self, record):
        d = record.msg
        if isinstance(d, dict):
            if d["event"] == "compilation":
                self.blocks.append({
                    "task": d["task"],
                    "source": d["source"],
                    "passes": d["passes"],
                    "emitted": d["emitted"],
                    "emit_time": d["emit_time"],
                    "results": None,
                    "elapsed": None
                })
            elif d["event"] == "time" and d["type"] == "query" and not d.get("internal"):
                last = self.blocks[-1]
                last["results"] = d["results"]
                last["elapsed"] = d["elapsed"]

################################################################################
# Snapshot validation
################################################################################

class SnapshotError(Exception):
    def __init__(self, file_path: Path, code: str, task: Task, header: str|None = None, footer: str|None = None):
        self.file_path = file_path
        self.source_code = code
        self.task = task
        self.header = header
        self.footer = footer

    def __str__(self):
        file_path = self.file_path

        with io.StringIO() as buf:
            console = rich.console.Console(file=buf, force_terminal=True)
            if self.header:
                console.print(self.header)

            source_info = debugging.get_source(self.task)
            assert source_info and source_info.line is not None
            source = debugging.find_block_in(self.source_code, source_info.line, str(file_path))

            base = os.getcwd()
            console.print("\nIn", f"./{file_path.relative_to(base)}" if file_path.is_relative_to(base) else file_path)
            if source.source:
                console.print(Syntax(source.source, "python", dedent=True, line_numbers=True, start_line=source.line, padding=1))

            if self.footer:
                console.print("\n" + self.footer)
            v = buf.getvalue()
        return v

    def __repr__(self):
        return self.__str__()

class ChangedSnapshotError(SnapshotError):
    def __init__(self, file_path: Path, code: str, task: Task, snap_file: str, diff: str):
        header = f"The value for the following block no longer matches its snapshot ({snap_file}):"
        footer = f"[bold]Diff[/bold]\n{diff}\n\n" + \
        "[yellow]If this change is expected, you can update the snapshot by running pytest with [bold]--snapshot-update[/bold][/yellow]"
        super().__init__(file_path, code, task, header, footer)

class NewSnapshotError(SnapshotError):
    def __init__(self, file_path: Path, code: str, task: Task, value: str|None):
        header = "No snapshot exists for the following block:"
        footer = textwrap.dedent(f"""
        Got:
        {value}
        [yellow]To save the current result as a snapshot, run pytest with [bold]--snapshot-update[/bold][/yellow]
        """)
        super().__init__(file_path, code, task, header, footer)


def exec_traced(code: str, model_kwargs:dict|None = None, file_path: str|None = None) -> list[Block]:
    # install logger to capture Rel compilation
    query_logger = QueryTextLogger()
    logger.addHandler(query_logger)


    model_kwargs = model_kwargs or {}
    if file_path:
        # ensure that the exec knows what file is being run
        model_kwargs["__file__"] = str(file_path)

    try:
        if file_path:
            code_object = compile(code, file_path, 'exec')
            exec(code_object, model_kwargs)
        else:
            exec(code, model_kwargs)
    except RAIException as err:
        err.pprint() # @FIXME: Is this still needed?
        raise err
    finally:
        logger.removeHandler(query_logger)

    return query_logger.blocks

def validate_block_snapshots(
        file_path: Path,
        snapshot,
        get_snapshot_str: Callable, # given a block, return the snapshot string
        snapshot_prefix:str,
        model_kwargs:dict|None = None
    ):
    """Execute the program at `file_path`, applying `get_snapshot_str` to all
    blocks and ensuring the result has not changed since the last run."""
    with open(file_path, "r") as file:
        code = file.read()
        # @TODO: Consider suppressing stdout

        blocks = exec_traced(code, model_kwargs, str(file_path))
        block_index = 0
        for block in blocks:
            snapshot_path = f"{snapshot_prefix}{block_index}.txt"
            cmp_string = None
            try:
                cmp_string = get_snapshot_str(block)
                if cmp_string is not None:
                    snapshot.assert_match(cmp_string, snapshot_path)
                    # only counting blocks that actually assert something
                    block_index += 1
            except RAIException as err:
                err.pprint()
                raise err from None
            except AssertionError as err:
                lines = str(err).splitlines()
                if "doesn't exist" in lines[0]:
                    raise NewSnapshotError(file_path, code, block["task"], cmp_string) from None

                if len(lines) >= 3 and "does not match" in lines[0]:
                    diff = '\n'.join(lines[3:])
                    raise ChangedSnapshotError(file_path, code, block["task"], snapshot_path, diff) from None

                raise


def randomize(name: str) -> str:
    return f"{name}_{random.randint(1000000000, 9999999999)}"

def do_snapshot_test(
        snapshot,
        file_path: Path,
        config: cfg.Config|None = None,
        engine_size: str|None = None,
        runs: int = 1,
        outer_span_type: str = "test",
        create_db: bool = True,
    ):
    """Execute the program at `file_path`, validating that query results
    have not changed since the last run. Uses the given `config` and
    `engine_size` if provided, or the ambient config if not."""
    test_name = file_path.stem

    provider = rai.Resources(config=config)
    config = provider.config

    def get_results(event: Block):
        if event["results"]:
            return str(event["results"]["values"])

    with debugging.span(outer_span_type, name=test_name, engine_size=engine_size):
        for i in range(runs):
            with debugging.span("run", idx=i):
                with debugging.span("program", main=os.path.relpath(file_path), platform=config.get("platform", None)) as span:
                    db_name = randomize(test_name)
                    validate_block_snapshots(file_path, snapshot, get_results, "query", {
                        "name": db_name,
                        "config": config,
                        "span": span,
                    })
                    # Some tests always use the same DB, never creating a DB named `db_name`,
                    # so we don't need to delete it.
                    if create_db:
                        try:
                            provider.delete_graph(db_name)
                        except Exception as e:
                            print(f"Failed to delete graph {db_name}: {e}")



from __future__ import annotations
from abc import ABC
from datetime import datetime
from typing import Any, Optional, TypedDict
from pathlib import Path
from urllib.parse import urlparse

class Import(TypedDict):
    id: str
    name: str
    type: str
    model: str
    hash: Optional[str]
    info: Optional[dict]
    status: Optional[str]

class AvailableModel(TypedDict):
    name: Any
    id: Any
    created_by: Any
    created_on: Any
    state: Any
    deleted_by: Any
    deleted_on: Any

class EngineState(TypedDict):
    name: str
    id: str
    size: str
    state: str
    created_by: str
    created_on: datetime
    updated_on: datetime

    version: str|None
    auto_suspend: int|None
    suspends_at: datetime|None

class SourceInfo(TypedDict):
    type: str|None
    state: str

class ImportSource(ABC):
    name: str

class ImportSourceFile(ImportSource):
    def __init__(self, path: str, name: Optional[str] = None):
        self.raw_path = path
        self._url = urlparse(self.raw_path)
        self._path = Path(self._url.path if self.is_url() else self.raw_path)
        self.extension = self._path.suffix
        self.name = name or self._path.name

    def is_url(self):
        return bool(self._url.netloc)

class ImportSourceTable(ImportSource):
    def __init__(self, database: str, schema: str, table: str, name: Optional[str] = None):
        self.database = database
        self.schema = schema
        self.table = table
        self.name = name or self.fqn

    @property
    def fqn(self):
        return f"{self.database}.{self.schema}.{self.table}"
    
class TransactionAsyncResponse:
    def __init__(
        self,
        transaction: dict|None = None,
        metadata: Any = None,
        results: list|None = None,
        problems: list|None = None,
    ):
        self.transaction = transaction
        self.metadata = metadata
        self.results = results
        self.problems = problems

    def __str__(self):
        return str(
            {
                "transaction": self.transaction,
                "metadata": self.metadata,
                "results": self.results,
                "problems": self.problems,
            }
        )

class ImportsStatus(TypedDict):
    engine: str
    engine_status: str
    engine_size: str
    status: str
    enabled: bool
    info: Any # @FIXME: This appears to be a stringified JSON (!?) object containing the same data as this dict + more

class SourceMapEntry(TypedDict):
    rel_end_line: int
    task_id: int
    py_line: int



from __future__ import annotations
from decimal import Decimal
from typing import Any, Union

from .. import dsl

_Number = Union[dsl.Producer, float, int, Decimal]

ConcreteReal = Union[int, float]

def isnumber(x: Any) -> bool:
    return isinstance(x, (int, float))

# NOTE: Right now, common contains all Rel stdlib relations.
# If the stdlib is split into multiple namespaces, this will have to be updated.
_math_ns = dsl.global_ns.std.common


# ------------------------------
# Basics
# ------------------------------

def abs(value: _Number) -> dsl.Expression:
    return _math_ns.abs(value)

def ceil(value: _Number) -> dsl.Expression:
    return _math_ns.ceil(value)

def clip(value: _Number, lower: _Number, upper: _Number) -> dsl.Expression:
    return _math_ns.clamp(lower, upper, value)

def factorial(value: _Number) -> dsl.Expression:
    if isnumber(value) and value < 0:
        raise ValueError("Cannot take the factorial of a negative number")
    return _math_ns.factorial(value)

def floor(value: _Number) -> dsl.Expression:
    return _math_ns.floor(value)

def isclose(x: _Number, y: _Number, tolerance: _Number = 1e-9) -> dsl.Expression:
    return _math_ns.approx_eq(tolerance, x, y)

def sign(x: _Number) -> dsl.Expression:
    return _math_ns.sign(x)

def trunc_divide(numerator: _Number, denominator: _Number) -> dsl.Expression:
    return _math_ns.trunc_divide(numerator, denominator)

# -------------------------------
# Power and Logarithmic Functions
# -------------------------------

def cbrt(value: _Number) -> dsl.Expression:
    return _math_ns.cbrt(value)

def exp(value: _Number) -> dsl.Expression:
    return _math_ns.natural_exp(value)

def log(x: _Number, base: _Number | None = None) -> dsl.Expression:
    if isnumber(x) and x <= 0:
        raise ValueError("Cannot take the logarithm of a non-positive number")
    if base is None:
        return _math_ns.natural_log(x)
    return _math_ns.log(base, x)

def log2(value: _Number) -> dsl.Expression:
    return log(value, 2)

def log10(value: _Number) -> dsl.Expression:
    return log(value, 10)

def pow(base: _Number, exponent: _Number) -> dsl.Expression:
    return _math_ns.power(base, exponent)

def sqrt(value: _Number) -> dsl.Expression:
    if isnumber(value) and value < 0:
        raise ValueError("Cannot take the square root of a negative number")
    return _math_ns.sqrt(value)


# ------------------------------
# Trigonometry
# ------------------------------

def degrees(radians: _Number) -> dsl.Expression:
    return _math_ns.rad2deg(radians)

def radians(degrees: _Number) -> dsl.Expression:
    return _math_ns.deg2rad(degrees)

def cos(value: _Number) -> dsl.Expression:
    return _math_ns.cos(value)

def sin(value: _Number) -> dsl.Expression:
    return _math_ns.sin(value)

def tan(value: _Number) -> dsl.Expression:
    return _math_ns.tan(value)

def cot(value: _Number) -> dsl.Expression:
    return _math_ns.cot(value)

def acos(value: _Number) -> dsl.Expression:
    return _math_ns.acos(value)

def asin(value: _Number) -> dsl.Expression:
    return _math_ns.asin(value)

def atan(value: _Number) -> dsl.Expression:
    return _math_ns.atan(value)

def acot(value: _Number) -> dsl.Expression:
    return _math_ns.acot(value)

def cosh(value: _Number) -> dsl.Expression:
    return _math_ns.cosh(value)

def sinh(value: _Number) -> dsl.Expression:
    return _math_ns.sinh(value)

def tanh(value: _Number) -> dsl.Expression:
    return _math_ns.tanh(value)

def acosh(value: _Number) -> dsl.Expression:
    if isnumber(value) and value < 1:
        raise ValueError("acosh expects a value greater than or equal to 1.")
    return _math_ns.acosh(value)

def asinh(value: _Number) -> dsl.Expression:
    return _math_ns.asinh(value)

def atanh(value: _Number) -> dsl.Expression:
    if isnumber(value) and (value <= -1 or value >= 1):
        raise ValueError("atanh expects a value between -1 and 1, exclusive.")
    return _math_ns.atanh(value)

# ------------------------------
# Special Functions
# ------------------------------

def erf(value: _Number) -> dsl.Expression:
    return _math_ns.erf(value)

def erfinv(value: _Number) -> dsl.Expression:
    if isnumber(value) and (value < -1 or value > 1):
        raise ValueError("erfinv expects a value between -1 and 1, inclusive.")
    return _math_ns.erfinv(value)

def haversine(x1: _Number, y1: _Number, x2: _Number, y2: _Number, radius: _Number = 1.0) -> dsl.Expression:
    return _math_ns.haversine(radius, x1, y1, x2, y2)


# ------------------------------
# Exports
# ------------------------------

__all__ = [
    "abs",
    "ceil",
    "clip",
    "factorial",
    "floor",
    "isclose",
    "sign",
    "trunc_divide",
    "cbrt",
    "exp",
    "log",
    "log2",
    "log10",
    "pow",
    "sqrt",
    "degrees",
    "radians",
    "cos",
    "sin",
    "tan",
    "cot",
    "acos",
    "asin",
    "atan",
    "acot",
    "cosh",
    "sinh",
    "tanh",
    "acosh",
    "asinh",
    "atanh",
    "erf",
    "erfinv",
    "haversine",
]



from __future__ import annotations
from collections import defaultdict
from copy import deepcopy
from decimal import Decimal
from numbers import Number, Real
import textwrap
from typing import Any, Type, Union

from ..errors import (
    DirectedGraphNotApplicable, DirectedGraphNotSupported,
    ParameterBoundBelowInclusive, ParameterBoundAboveInclusive,
    ParameterBoundBelowExclusive, ParameterBoundAboveExclusive,
    ParameterTypeMismatch, UnsupportedVisualizationError,
)
from .. import dsl
from ..environments import runtime_env, SnowbookEnvironment
from . import rel, Vars

Numeric = Union[int, float, Decimal]
NumericType = Type[Union[Numeric, Number]]

#--------------------------------------------------
# Standardized input validation functions
#--------------------------------------------------

def assert_type(name: str, value: Numeric, type_: NumericType):
    if not isinstance(value, type_):
        raise ParameterTypeMismatch(name, type_, value)

def assert_inclusive_lower_bound(name: str, value: Numeric, minimum: Numeric):
    if value < minimum:
        raise ParameterBoundBelowInclusive(name, value, minimum)

def assert_inclusive_upper_bound(name: str, value: Numeric, maximum: Numeric):
    if value > maximum:
        raise ParameterBoundAboveInclusive(name, value, maximum)

def assert_exclusive_lower_bound(name: str, value: Numeric, minimum: Numeric):
    if value <= minimum:
        raise ParameterBoundBelowExclusive(name, value, minimum)

def assert_exclusive_upper_bound(name: str, value: Numeric, maximum: Numeric):
    if value >= maximum:
        raise ParameterBoundAboveExclusive(name, value, maximum)


#--------------------------------------------------
# Helpers
#--------------------------------------------------

def unwrap(v):
    if v is None:
        return None
    return rel.pyrel_default(rel.pyrel_unwrap, v, v)


#--------------------------------------------------
# Algos
#--------------------------------------------------

class Compute:
    def __init__(self, graph:'Graph'):
        self._graph = graph
        self._lib = dsl.global_ns.graphlib

    def _config(self, **kwargs):
        return dsl.InlineRelation(self._graph.model, [
            *[(dsl.Symbol(k), v) for k, v in kwargs.items()],
        ])

    def _lookup(self, v):
        res = dsl.create_var()
        lookup_rel = getattr(rel, f"{self._graph._graph_ref()}_lookup")
        rel.pyrel_default(lookup_rel, v, v, res)
        return res

    # --------------------------------------------------
    # Degree
    # --------------------------------------------------

    def degree(self, node, weight=None):
        if not weight:
            return self._lib.degree(self._graph, unwrap(node))
        return self._lib.weighted_degree(self._graph, unwrap(node))

    def indegree(self, node):
        return self._lib.indegree(self._graph, unwrap(node))

    def outdegree(self, node):
        return self._lib.outdegree(self._graph, unwrap(node))

    # --------------------------------------------------
    # Basics
    # --------------------------------------------------

    def num_nodes(self):
        return self._lib.num_nodes(self._graph)

    def num_edges(self):
        return self._lib.num_edges(self._graph)

    # --------------------------------------------------
    # Distance
    # --------------------------------------------------

    def distance(self, node1, node2):
        if self._graph.weighted:
            return self._lib.weighted_distance(self._graph, unwrap(node1), unwrap(node2))
        return self._lib.distance(self._graph, unwrap(node1), unwrap(node2))

    def diameter_range(self):
        min, max = dsl.create_vars(2)
        self._lib.diameter_range(self._graph, "min", min)
        self._lib.diameter_range(self._graph, "max", max)
        return (min, max)

    # --------------------------------------------------
    # Connectivity
    # --------------------------------------------------

    def is_connected(self):
        return self._lib.is_connected(self._graph)

    def reachable_from(self, node: dsl.Producer) -> dsl.Producer:
        reachable = self._lib.reachable_from(self._graph, unwrap(node))
        return self._lookup(reachable)

    def is_reachable(self, node1: dsl.Producer, node2: dsl.Producer) -> dsl.Expression:
        return self._lib.reachable_from(self._graph, unwrap(node1), unwrap(node2))

    # --------------------------------------------------
    # Triangles
    # --------------------------------------------------

    def num_triangles(self, node=None):
        if node is None:
            return self._lib.num_triangles(self._graph)
        return self._lib.triangle_count(self._graph, unwrap(node))

    def is_triangle(self, node1, node2, node3):
        return self._lib.triangle(self._graph, unwrap(node1), unwrap(node2), unwrap(node3))

    def triangles(self, node=None):
        # If no node is provided, return all triangles in the graph,
        # unique up to ordering of the nodes.
        if node is None:
            x, y, z = dsl.create_vars(3)
            self._lib.unique_triangle(self._graph, x, y, z)
            lx = self._lookup(x)
            ly = self._lookup(y)
            lz = self._lookup(z)
            return (lx, ly, lz)
        # If a node is provided, return all triangles that include that node,
        # unique up to order of the nodes.
        x, y = dsl.create_vars(2)
        model = dsl.get_graph()
        node = unwrap(node)
        with model.union(dynamic=True) as t:
            for (a, b, c) in ([node, x, y], [x, node, y], [x, y, node]):
                with model.scope():
                    self._lib.unique_triangle(self._graph, a, b, c)
                    t.add(node, a=a, b=b, c=c)
        la = self._lookup(t.a)
        lb = self._lookup(t.b)
        lc = self._lookup(t.c)
        return (la, lb, lc)

    # --------------------------------------------------
    # Clustering
    # --------------------------------------------------

    def local_clustering_coefficient(self, node):
        if not self._graph.undirected:
            raise DirectedGraphNotApplicable("local_clustering_coefficient")
        return self._lib.local_clustering_coefficient(self._graph, unwrap(node))

    def avg_clustering_coefficient(self):
        if not self._graph.undirected:
            raise DirectedGraphNotApplicable("avg_clustering_coefficient")
        return self._lib.average_clustering_coefficient(self._graph)

    # --------------------------------------------------
    # Link Prediction
    # --------------------------------------------------

    def adamic_adar(self, node1, node2):
        return self._lib.adamic_adar(self._graph, unwrap(node1), unwrap(node2))

    def preferential_attachment(self, node1, node2):
        return self._lib.preferential_attachment(self._graph, unwrap(node1), unwrap(node2))

    def common_neighbor(self, node1, node2):
        neighbor = self._lib.common_neighbor(self._graph, unwrap(node1), unwrap(node2))
        return self._lookup(neighbor)

    # --------------------------------------------------
    # Similarity
    # --------------------------------------------------

    def cosine_similarity(self, node1, node2):
        if self._graph.weighted:
            return self._lib.weighted_cosine_similarity(self._graph, unwrap(node1), unwrap(node2))
        return self._lib.cosine_similarity(self._graph, unwrap(node1), unwrap(node2))

    def jaccard_similarity(self, node1, node2):
        if self._graph.weighted:
            return self._lib.weighted_jaccard_similarity(self._graph, unwrap(node1), unwrap(node2))
        return self._lib.jaccard_similarity(self._graph, unwrap(node1), unwrap(node2))

    # --------------------------------------------------
    # Centrality
    # --------------------------------------------------

    def pagerank(self, node, damping_factor=0.85, tolerance=1e-6, max_iter=20):
        assert_type("pagerank:tolerance", tolerance, Real)
        assert_exclusive_lower_bound("pagerank:tolerance", tolerance, 0.0)

        assert_type("pagerank:max_iter", max_iter, int)
        assert_exclusive_lower_bound("pagerank:max_iter", max_iter, 0)

        assert_type("pagerank:damping_factor", damping_factor, Real)
        assert_inclusive_lower_bound("pagerank:damping_factor", damping_factor, 0.0)
        assert_exclusive_upper_bound("pagerank:damping_factor", damping_factor, 1.0)

        config = self._config(
            graph=self._graph,
            damping_factor=float(damping_factor),
            tolerance=float(tolerance),
            max_iter=max_iter,
        )

        return self._lib.pagerank(config, unwrap(node))

    def betweenness_centrality(self, node):
        return self._lib.betweenness_centrality(self._graph, unwrap(node))

    def degree_centrality(self, node):
        if self._graph.weighted:
            return self._lib.weighted_degree_centrality(self._graph, unwrap(node))
        return self._lib.degree_centrality(self._graph, unwrap(node))

    def eigenvector_centrality(self, node):
        if not self._graph.undirected:
            raise DirectedGraphNotSupported("eigenvector_centrality")
        return self._lib.eigenvector_centrality(self._graph, unwrap(node))


    # --------------------------------------------------
    # Community Detection
    # --------------------------------------------------

    def weakly_connected_component(self, node):
        component = self._lib.weakly_connected_component(self._graph, unwrap(node))
        return self._lookup(component)

    def triangle_community(self, node):
        if not self._graph.undirected:
            raise DirectedGraphNotApplicable("triangle_community")
        return self._lib.triangle_community(self._graph, unwrap(node))

    def infomap(
            self,
            node,
            max_levels: int = 1,
            max_sweeps: int = 20,
            level_tolerance: float = 0.01,
            sweep_tolerance: float = 0.0001,
            teleportation_rate: float = 0.15,
            visit_rate_tolerance: float = 1e-15,
            randomization_seed: int | None = None,
    ):
        assert_type("infomap:max_levels", max_levels, int)
        assert_type("infomap:max_sweeps", max_sweeps, int)
        assert_exclusive_lower_bound("infomap:max_levels", max_levels, 0)
        assert_inclusive_lower_bound("infomap:max_sweeps", max_sweeps, 0)

        assert_type("infomap:level_tolerance", level_tolerance, Real)
        assert_type("infomap:sweep_tolerance", sweep_tolerance, Real)
        assert_inclusive_lower_bound("infomap:level_tolerance", level_tolerance, 0.0)
        assert_inclusive_lower_bound("infomap:sweep_tolerance", sweep_tolerance, 0.0)

        assert_type("infomap:teleportation_rate", teleportation_rate, Real)
        assert_inclusive_lower_bound("infomap:teleportation_rate", teleportation_rate, 1e-4)
        assert_exclusive_upper_bound("infomap:teleportation_rate", teleportation_rate, 1.0)

        assert_type("infomap:visit_rate_tolerance", visit_rate_tolerance, Real)
        assert_exclusive_lower_bound("infomap:visit_rate_tolerance", visit_rate_tolerance, 0.0)

        _config_dict = {
            "graph": self._graph,
            "max_levels": max_levels,
            "max_sweeps": max_sweeps,
            "level_tolerance": float(level_tolerance),
            "sweep_tolerance": float(sweep_tolerance),
            "teleportation_rate": float(teleportation_rate),
            "visit_rate_tolerance": float(visit_rate_tolerance),
        }

        if randomization_seed is not None:
            assert_type("infomap:randomization_seed", randomization_seed, int)
            assert_exclusive_lower_bound("infomap:randomization_seed", randomization_seed, 0)
            _config_dict["randomization_seed"] = randomization_seed

        config = self._config(**_config_dict)
        return self._lib.infomap(config, unwrap(node))

    def louvain(
            self,
            node,
            max_levels: int = 1,
            max_sweeps: int = 20,
            level_tolerance: float = 0.01,
            sweep_tolerance: float = 0.0001,
            randomization_seed: int | None = None,
    ):
        if not self._graph.undirected:
            raise DirectedGraphNotApplicable("louvain")

        assert_type("louvain:max_levels", max_levels, int)
        assert_type("louvain:max_sweeps", max_sweeps, int)
        assert_exclusive_lower_bound("louvain:max_levels", max_levels, 0)
        assert_inclusive_lower_bound("louvain:max_sweeps", max_sweeps, 0)

        assert_type("louvain:level_tolerance", level_tolerance, Real)
        assert_type("louvain:sweep_tolerance", sweep_tolerance, Real)
        assert_inclusive_lower_bound("louvain:level_tolerance", level_tolerance, 0.0)
        assert_inclusive_lower_bound("louvain:sweep_tolerance", sweep_tolerance, 0.0)

        _config_dict = {
            "graph": self._graph,
            "max_levels": max_levels,
            "max_sweeps": max_sweeps,
            "level_tolerance": float(level_tolerance),
            "sweep_tolerance": float(sweep_tolerance),
        }

        if randomization_seed is not None:
            assert_type("louvain:randomization_seed", randomization_seed, int)
            assert_exclusive_lower_bound("louvain:randomization_seed", randomization_seed, 0)
            _config_dict["randomization_seed"] = randomization_seed

        config = self._config(**_config_dict)
        return self._lib.louvain(config, unwrap(node))

    def label_propagation(self, node, max_sweeps: int = 20, randomization_seed: int | None = None):
        assert_type("label_propagation:max_sweeps", max_sweeps, int)
        assert_inclusive_lower_bound("label_propagation:max_sweeps", max_sweeps, 0)

        _config_dict = {
            "graph": self._graph,
            "max_sweeps": max_sweeps,
        }

        if randomization_seed is not None:
            assert_type("label_propagation:randomization_seed", randomization_seed, int)
            assert_exclusive_lower_bound("label_propagation:randomization_seed", randomization_seed, 0)
            _config_dict["randomization_seed"] = randomization_seed

        config = self._config(**_config_dict)
        return self._lib.label_propagation(config, unwrap(node))

#--------------------------------------------------
# Edge
#--------------------------------------------------

class EdgeInstance:
    def __init__(self, edge:'Edge', from_:Any, to:Any, kwargs:dict={}):
        self._edge = edge
        self._graph = edge._graph
        self.from_ = from_
        self.to = to
        for k, v in kwargs.items():
            self._edge._prop(k)(from_, to, v)

    def __getattr__(self, name:str):
        v = Vars(1)
        self._edge._prop(name)(self.from_, self.to, v)
        return v

    def set(self, **kwargs):
        for k, v in kwargs.items():
            self._edge._prop(k).add(self.from_, self.to, v)
        return self

    def _to_var(self):
        raise Exception("Edges can't be returned directly, you can return the from_ and to properties individually")

class Edge:
    def __init__(self, graph:'Graph'):
        self._graph = graph
        self._type = dsl.RawRelation(self._graph.model, f"graph{self._graph.id}_edges", 2)
        self._props = {}

    def _prop(self, name:str):
        if name not in self._props:
            self._props[name] = dsl.RawRelation(self._graph.model, f"graph{self._graph.id}_edge_{name}", 3)
        return self._props[name]

    def _is_weighted(self):
        return "weight" in self._props

    def extend(self, prop:'dsl.Property', **kwargs):
        type = prop._provider
        with self._graph.model.rule():
            t = type()
            self.add(t, getattr(t, prop._name), **kwargs)

    def add(self, from_:Any, to:Any, **kwargs):
        self._type.add(from_, to)
        for k, v in kwargs.items():
            self._prop(k).add(from_, to, v)
        return EdgeInstance(self, from_, to, {})

    def __call__(self, from_:Any=None, to:Any=None, **kwargs):
        if from_ is None:
            from_ = Vars(1)
        if to is None:
            to = Vars(1)
        self._type(from_, to)
        return EdgeInstance(self, from_, to, kwargs)

    def __getattribute__(self, __name: str) -> Any:
        if __name in ["add", "extend"] or __name.startswith("_"):
            return super().__getattribute__(__name)
        return self._props[__name]

#--------------------------------------------------
# Graph
#--------------------------------------------------

class Graph:
    def __init__(self, model:dsl.Graph, undirected=False, weighted=False, default_weight=1.0, with_isolated_nodes=True):
        self.model = model
        self.id = dsl.next_id()
        self.compute = Compute(self)
        self.Node = dsl.Type(model, "nodes", scope=f"graph{self.id}_", omit_intrinsic_type_in_hash=True)
        self.Edge = Edge(self)
        self._undirected = undirected
        self._weighted = weighted
        self.default_weight=default_weight
        self._last_fetch = None

        create_graph_lines = []
        graph_type = "::graphlib::undirected_graph" if undirected else "::graphlib::directed_graph"
        if weighted:
            unwrapped = f"""
            def {self.Edge._prop("weight")._name}_unwrapped(au, bu, w): {{
                exists((a, b) |
                    {self.Edge._prop("weight")._name}(a, b, w) and
                    pyrel_default(pyrel_unwrap, a, a, au) and
                    pyrel_default(pyrel_unwrap, b, b, bu)
                )
            }}
            """
            create_graph_lines.extend([
                f"""{graph_type}[{{""",
                f"""  (:edge, {self.Edge._prop("weight")._name}_unwrapped);"""
            ])
        else:
            unwrapped = f"""
            def {self.Edge._type._name}_unwrapped(au, bu): {{
                exists((a, b) |
                    {self.Edge._type._name}(a, b) and
                    pyrel_default(pyrel_unwrap, a, a, au) and
                    pyrel_default(pyrel_unwrap, b, b, bu)
                )
            }}
            """
            create_graph_lines.extend([
                f"""{graph_type}[{{""",
                f"""  (:edge, {self.Edge._type._name}_unwrapped);"""
            ])

        if with_isolated_nodes:
            create_graph_lines.append(
                f"  (:node, pyrel_unwrap[{self.Node._type.name}]);"
            )
        create_graph_lines.extend([
            "  (:diagnostics)",
            "}]"
        ])
        create_graph = "\n".join(create_graph_lines)

        self.model.install_raw(textwrap.dedent(f"""
        declare {self.Node._type.name}
        declare {self.Edge._type._name}
        {unwrapped}
        @inline
        def {self._graph_ref()}_lookup(uw, orig): {{ {self.Node._type.name}(orig) and pyrel_unwrap(orig, uw) }}

        {f"declare {self.Edge._prop('weight')._name}" if weighted else ""}
        def {self._graph_ref()} {{{create_graph}}}
        def {self._graph_result_ref()} {{{self._graph_ref()}[:result]}}
        ic graphlib_edge_error(x...) requires not {self._graph_ref()}(:error, x...)
        """))

        # Add a rule that makes all nodes used in edges are also added to
        # the nodes relation
        with model.rule():
            a, b = dsl.create_vars(2)
            self.Edge._type(a, b)
            self.Node.add(a)
            self.Node.add(b)

    def _graph_ref(self):
        return f"graph{self.id}"

    def _graph_result_ref(self):
        return f"graph{self.id}_result"

    def _to_var(self):
        return getattr(rel, self._graph_result_ref())._to_var()

    def _is_weighted(self):
        return self.weighted
        # return self.edges._is_weighted()

    @property
    def undirected(self):
        return self._undirected

    @property
    def weighted(self):
        return self._weighted

    #--------------------------------------------------
    # Fetch
    #--------------------------------------------------

    def fetch(self):
        code = []
        code.append(f"def output(:nodes, n): {self.Node._type.name}(n)")
        for prop in self.Node._type.properties:
            scope = self.Node._scope
            code.append(f"def output(:nodes, :{prop.name.removeprefix(scope)}, n, v): {prop.name}(n, v)")
        code.append(f"def output(:edges, a,b): {self.Edge._type._name}(a,b)")
        for name, prop in self.Edge._props.items():
            code.append(f"def output(:edges, :{name}, a, b, v): {prop._name}(a, b, v)")

        output = {"nodes": defaultdict(dict), "edges": defaultdict(dict)}
        results = self.model.exec_raw("\n".join(code), raw_results=True)
        for set_ in results.results:
            path = [v[1:] for v in set_["relationId"].split("/")[2:] if v[0] == ":"]
            if path[0] not in ["nodes", "edges"]:
                continue
            cur = output[path[0]]
            if path[0] == "nodes":
                if len(path) == 1:
                    for (n,) in set_["table"].itertuples(index=False):
                        cur[n]
                else:
                    for (n, v) in set_["table"].itertuples(index=False):
                        cur[n][path[1]] = v
            elif path[0] == "edges":
                if len(path) == 1:
                    for (a, b) in set_["table"].itertuples(index=False):
                        # If the graph is undirected, normalize the order of
                        # the source and target nodes.
                        if self.undirected and a > b:
                            a, b = b, a
                        cur[(a, b)]
                else:
                    for (a, b, v) in set_["table"].itertuples(index=False):
                        if self.undirected and a > b:
                            a, b = b, a
                        cur[(a, b)][path[1]] = v
            else:
                raise Exception(f"Unexpected path: {path}")
        self._last_fetch = output
        return output

    #--------------------------------------------------
    # Visualize
    #--------------------------------------------------

    default_visual_props = {
        "node": {
            "color": "black",
            "opacity": 1,
            "size": 10,
            "shape": "circle",
            "border_color": "black",
            "border_size": 1,
            "label_color": "black",
            "label_size": 10,
        },
        "edge": {
            "color": "#999",
            "opacity": 1,
            "size": 2,
            "shape": "circle",
            "border_color": "#999",
            "border_size": 1,
            "label_color": "black",
            "label_size": 10,
            "arrow_size": 4,
            "arrow_color": "#999",
        }
    }

    _style_map = {
        "arrow_size": "arrow_size",
        "arrow_color": "arrow_color",
    }

    def _visual_props(self, prop_def, metadata):
        for k, v in prop_def.items():
            if callable(v):
                metadata[k] = v(metadata)
        # For some reason, the existance of "id" in the metadata
        # results in edges not getting displayed in the visualization,
        # so we remove it.
        if "id" in metadata:
            del metadata["id"]
        return metadata

    def _visual_dict(self, style: dict, use_cache = False) -> dict:
        data = self._last_fetch if use_cache else None
        if not data:
            data = self.fetch()
        style = deepcopy(style)
        if "node" not in style:
            style["node"] = {}
        if "edge" not in style:
            style["edge"] = {}
        merged_style = deepcopy(self.default_visual_props)
        for category in ["node", "edge"]:
            for k, v in style.get(category, {}).items():
                if not callable(v):
                    merged_style[category][k] = v
        flat_style = {
            self._style_map.get(k, f"{type}_{k}"): v
            for type, category in merged_style.items()
            for k, v in category.items()
        }

        return {
            "graph": {
                "directed": not self.undirected,
                "metadata": flat_style,
                "nodes": {
                    node_id: {
                        **({"label": info["label"]} if "label" in info else {}),
                        "metadata": self._visual_props(style["node"], info),
                    }
                    for (node_id, info) in data["nodes"].items()
                },
                "edges": [
                    {"source": source, "target": target, "metadata": self._visual_props(style["edge"], info)}
                    for ((source, target), info) in data["edges"].items()
                ],
            }
        }

    def visualize(self, three=False, style: dict | None = None, use_cache = False, **kwargs):
        if isinstance(runtime_env, SnowbookEnvironment):
            raise UnsupportedVisualizationError()
        import gravis as gv
        style = style if style is not None else {"node": {}, "edge": {}}
        vis = gv.vis if not three else gv.three
        graph_dict = self._visual_dict(style=style, use_cache=use_cache)
        # Use defaults for the following kwargs if not provided
        new_kwargs = {
            "node_label_data_source": "label",
            "edge_label_data_source": "label",
            "show_edge_label": True,
            "edge_curvature": 0.4,
        } | kwargs
        fig = vis(graph_dict, **new_kwargs)
        return fig

#--------------------------------------------------
# Path
#--------------------------------------------------

class Path:
    def __init__(self, *args):
        self.edges = []
        self.nodes = []
        pass

    def __getitem__(self, item):
        return self

#--------------------------------------------------
# Exports
#--------------------------------------------------

__all__ = ["Graph", "Path"]



import hashlib

def hash_string(s: str, digits=32):
    return hashlib.md5(s.encode()).hexdigest()[:digits]

def database_name_from_sproc_name(sproc_name: str) -> str:
    if sproc_name.count(".") != 2:
        raise ValueError(f"Invalid stored procedure name: {sproc_name}")
    if len(sproc_name) < 32:
        return f"{sproc_name.replace('.', '_')}_{hash_string(sproc_name)}"
    database, schema, proc = sproc_name.split(".")
    database = database[:10]
    schema = schema[:10]
    proc = proc[:(31 - len(database) - len(schema))]
    hash = hash_string(sproc_name)
    return f"{database}_{schema}_{proc}_{hash}"


from __future__ import annotations
import os
import sys
import warnings

from .base import RuntimeEnvironment, SourceInfo, find_external_frame, handle_rai_exception, handle_rai_warning, patch

class GenericEnvironment(RuntimeEnvironment):
    @classmethod
    def detect(cls):
        return True

    def get_source(self, steps:int|None = None):
        caller_frame = find_external_frame(steps)
        if not caller_frame:
            return

        caller_filename = caller_frame.f_code.co_filename
        caller_line = caller_frame.f_lineno

        relative_filename = os.path.relpath(caller_filename, os.getcwd())

        source_code = None
        try:
            # Read the source code from the caller's file
            with open(caller_filename, "r") as f:
                source_code = f.read()
        except Exception:
            pass

        return SourceInfo.from_source(relative_filename, caller_line, source_code)

    def _patch(self):
        @patch(warnings, "showwarning")
        def _(original, message, category, filename, lineno, file=None, line=None):
            if not handle_rai_warning(message):
                original(message, category, filename, lineno, file, line)

        @patch(sys, "excepthook")
        def _(original, exc_type, exc_value, exc_traceback, quiet=False):
            handle_rai_exception(exc_value)
            original(exc_type, exc_value, exc_traceback)
